import itertools
import math
from datetime import date, datetime
from itertools import permutations

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from profiling.shap import ShapExplainer, comb, n_choose_k, reduce_func, update_columns
from profiling.utils import batched_non_object_col_mode, batched_numeric_col_mean


@pytest.mark.parametrize("n_cols", [1, 2, 3, 4])
def test_coalitions(n_cols):
    """Test construction of coalitions"""
    np.random.seed(1986)
    X = pd.DataFrame(np.random.normal(size=(1, n_cols)))

    def func(x):
        coef = np.ones(len(x))
        return x @ coef

    shap_explainer = ShapExplainer(func, X)
    coalitions = shap_explainer._sample_coalitions(len(X.columns))
    coalitions_expected = set(permutations([False, True] * (n_cols - 1), r=n_cols))
    assert coalitions == coalitions_expected


@pytest.mark.parametrize("coef", [[0], [1], [0.25, 0.5, 0, 0]])
def test_coefficient_recovery(coef):
    """Test that SHAP converges to coefficients in linear model"""
    np.random.seed(1986)
    X = pd.DataFrame(np.zeros(shape=(1, len(coef))))
    intercept = 5.23

    def func(X):
        return X.values @ coef + intercept

    shap_explainer = ShapExplainer(func, X)
    test_row = pd.DataFrame(np.ones((1, len(X.columns))), columns=X.columns)
    ref, shap = shap_explainer.explain(test_row)
    assert np.allclose(ref, intercept)  # reference
    assert np.allclose(shap, coef)  # shapley values
    assert np.allclose(ref + shap.sum(axis=1), func(test_row))  # bookkeeping


def test_large_df():
    """Test that SHAP runs smoothly on large dataframes"""
    np.random.seed(1986)
    n_col = 10
    n_row = 10000
    coef = np.random.randn(n_col)
    X = pd.DataFrame(np.zeros((n_row, n_col)))
    intercept = 5.23

    def func(X):
        return X.values @ coef + intercept

    shap_explainer = ShapExplainer(func, X)
    test_row = pd.DataFrame(np.ones((1, len(X.columns))), columns=X.columns)
    ref, shap = shap_explainer.explain(test_row)
    assert np.allclose(ref, intercept)  # reference
    assert np.allclose(shap, coef)  # shapley values
    assert np.allclose(ref + shap.sum(axis=1), func(test_row))  # bookkeeping


def test_batch():
    """Test that SHAP works in batch mode"""
    np.random.seed(1986)
    coef = [0.5, 0.5, 0, 0]
    X = pd.DataFrame(np.random.randn(10000, len(coef)))
    intercept = 5.23

    def func(X):
        return X.values @ coef + intercept

    shap_explainer = ShapExplainer(func, X)
    n_row = 10
    test_df = X.iloc[:n_row]
    ref, shap = shap_explainer.explain(test_df)
    assert isinstance(ref, pd.Series)
    assert isinstance(shap, pd.DataFrame)
    assert ref.shape == (n_row,)
    assert shap.shape == (n_row, len(coef))
    assert np.allclose(ref + shap.sum(axis=1), func(test_df))  # bookkeeping


def test_various_coltypes():
    X = pd.DataFrame.from_dict(
        {
            "Integer": [1, 2, 3],
            "Float": [None, 2.0, 3.0],
            "Object": [None, "", "c"],
            "Categorical": ["a", "b", "c"],
            "Date": [date(2020, 1, 1), date(2020, 1, 2), date(2020, 1, 3)],
            "Datetime": [datetime(2020, 1, 1), None, datetime(2020, 1, 3)],
            "Nullable Int64": pd.array([1, 3, None], dtype="Int64"),
        }
    )
    X = X.astype({"Categorical": "category"})

    def regression(x: pd.DataFrame) -> pd.Series:
        return pd.Series(np.random.uniform(size=len(x)))

    shap_explainer = ShapExplainer(func=regression, x=X)
    _, _ = shap_explainer.explain(X)


def test_comb():
    """Test n_choose_k and comb function"""
    assert n_choose_k(1, 1) == 1
    assert n_choose_k(2, 1) == 2
    assert n_choose_k(2, 2) == 1
    assert n_choose_k(4, 2) == 6
    assert n_choose_k(100, 50) == 100891344545564193334812497256
    assert np.array_equal(comb(3, np.array([1, 2, 3])), np.array([3, 3, 1]))


def test_background_data():
    """Test background data computation"""
    data = pd.DataFrame(
        {
            "Float": [1, 2.1, 2.3],
            "Repeated Values": [1, 1, 1],
            "Float with Nones": [None, 1.0, None],
            "Float with NaNs": [float("NaN"), 1.0, None],
            "Object": ["a", "c", "c"],
            "Object with Nones": [None, None, "c"],
            "Only NaNs": [float("NaN"), float("NaN"), float("NaN")],
            "Nullable Int64": pd.array([1, 3, None], dtype="Int64"),
            "Nullable Int8": pd.array([1, 3, None], dtype="Int8"),
            "Empty Nullable": pd.array([None, None, None], dtype="Int64"),
        }
    )

    expected = pd.DataFrame(
        {
            "Float": [1.8],
            "Repeated Values": [1],
            "Float with Nones": [1.0],
            "Float with NaNs": [1.0],
            "Object": ["c"],
            "Object with Nones": ["c"],
            "Only NaNs": [float("NaN")],
            "Nullable Int64": pd.array([2], dtype="Int64"),
            "Nullable Int8": pd.array([2], dtype="Int8"),
            "Empty Nullable": pd.array([pd.NA], dtype="Int64"),
        }
    )
    x_batched_init, x_batched = itertools.tee((data for _ in range(4)), 2)
    x_full = pd.concat([data for _ in range(4)])

    y_batched_init, y_batched = itertools.tee(
        (pd.Series([1, 1, 1]) for _ in range(4)), 2
    )

    full_explainer = ShapExplainer(func=lambda x: x, x=x_full)
    batched_explainer = ShapExplainer(
        func=lambda x: x, x=x_batched_init, y=y_batched_init
    )
    result_from_x_full = full_explainer._create_background_data(x_full)
    result_from_x_batched = batched_explainer._create_background_data(
        x_batched, y=y_batched
    )

    assert_frame_equal(result_from_x_full, expected)
    assert_frame_equal(result_from_x_batched, expected)


def test_mapping():
    X = pd.DataFrame.from_dict({"Integer": [3, 2, 2], "Categorical": ["a", "a", "b"]})
    X = X.astype({"Categorical": "category"})

    def regression(x):
        return pd.Series(np.random.uniform(size=len(x)))

    shap_explainer = ShapExplainer(func=regression, x=X)
    test_row = pd.DataFrame({"Integer": [5], "Categorical": ["d"]}).astype(X.dtypes)

    # full coalition
    coalition = {(True, True)}
    mapped_features = shap_explainer._map_features(coalition, test_row)
    expected = test_row
    assert mapped_features.equals(expected)

    # empty coalition
    coalition = {(False, False)}
    mapped_features = shap_explainer._map_features(coalition, test_row)
    expected = shap_explainer.x  # background dataset
    assert mapped_features.equals(expected)

    # mixed coalitions
    coalition = {(False, True)}
    mapped_features = shap_explainer._map_features(coalition, test_row)
    expected = pd.DataFrame({"Integer": [2], "Categorical": ["d"]}).astype(X.dtypes)
    assert mapped_features.equals(expected)

    coalition = {(True, False)}
    mapped_features = shap_explainer._map_features(coalition, test_row)
    expected = pd.DataFrame({"Integer": [5], "Categorical": ["a"]}).astype(X.dtypes)
    assert mapped_features.equals(expected)


def test_axioms():
    """Test Missingness, Local Accuracy, and Consistency properties of Shapley values"""
    np.random.seed(1986)
    n_col = 3
    X = pd.DataFrame(np.random.normal(size=(10000, n_col)))

    def func0(x):
        return x.loc[:, 0] * x.loc[:, 1] ** 2

    def func1(x):
        return x.loc[:, 0] * x.loc[:, 1] ** 2 + x.loc[:, 0].abs()

    explainer0 = ShapExplainer(func0, X)
    explainer1 = ShapExplainer(func1, X)

    test_row = pd.DataFrame(np.ones(shape=(1, n_col)))
    ref0, shap0 = explainer0.explain(test_row)
    ref1, shap1 = explainer1.explain(test_row)

    # missingness
    assert np.isclose(shap0[2], 0)
    assert np.isclose(shap1[2], 0)

    # local accuracy
    assert np.isclose(ref0 + np.sum(shap0, axis=1), func0(test_row))
    assert np.isclose(ref1 + np.sum(shap1, axis=1), func1(test_row))

    # consistency
    assert np.all(shap0[0] < shap1[0])
    assert np.all(shap0 < shap1 + 1e-9)


def test_profile_columns():
    """Test Shapley value calculation for a subset of columns"""
    np.random.seed(1986)
    n_col = 10
    n_row = 1000
    coef = np.random.randn(n_col)
    X = pd.DataFrame(np.zeros(shape=(n_row, n_col)))
    intercept = 5.23

    def func(x):
        return x.values @ coef + intercept

    profile_columns = X.columns[:-1]
    shap_explainer = ShapExplainer(func, X, profile_columns=profile_columns)
    test_row = pd.DataFrame(
        np.ones(shape=(1, len(profile_columns))), columns=profile_columns
    )
    ref, shap = shap_explainer.explain(test_row)
    assert np.allclose(ref, intercept)  # reference
    assert np.allclose(shap, coef[:-1])  # shapley values
    assert np.allclose(
        ref + shap.sum(axis=1), shap_explainer.func(test_row)
    )  # bookkeeping


def test_update_columns():
    # base case
    base = pd.DataFrame({"a": [1]})
    update = pd.DataFrame({"a": [2]})
    expected = pd.DataFrame({"a": [2]})
    result = update_columns(base, update)
    assert result.equals(expected)

    # broadcasting
    base = pd.DataFrame({"a": [1], "b": [True]})
    update = pd.DataFrame({"a": [2, 3]})
    expected = pd.DataFrame({"a": [2, 3], "b": [True, True]})
    result = update_columns(base, update)
    assert result.equals(expected)

    # index mismatch handled gracefully
    base = pd.DataFrame({"a": [1], "b": [True]}, index=[0])
    update = pd.DataFrame({"a": [2, 3]}, index=[1, 2])
    expected = pd.DataFrame({"a": [2, 3], "b": [True, True]}, index=[1, 2])
    result = update_columns(base, update)
    assert result.equals(expected)

    # column mismatch
    with pytest.raises(KeyError):
        base = pd.DataFrame({"a": [1]})
        update = pd.DataFrame({"b": [2, 3]})
        result = update_columns(base, update)


def test_reduce_func():
    X_orig = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    X_subset = X_orig.drop(columns=["a"])

    def predict(x: pd.DataFrame) -> pd.Series:
        assert list(x.columns) == ["a", "b"]
        return pd.Series([0] * len(x))

    with pytest.raises(AssertionError):
        predict(X_subset)

    predict_subset = reduce_func(func=predict, x=X_orig)
    pred = predict_subset(X_subset)
    pd.testing.assert_series_equal(pred, pd.Series([0, 0]))


def test_batched_means():
    x_data = [
        {"a": [1, 3], "b": [3, None]},
        {"a": [1, 3, 1, 0], "b": [3, 2.0, None, 1]},
        {"a": [1, 9], "b": [7, 7]},
    ]
    y_data = [[1, 3, 1, 0], [1, 3]]

    # duplicate iterators
    batched_x_a, batched_x_b = itertools.tee((pd.DataFrame(x) for x in x_data[1:]), 2)
    batched_y_a, batched_y_b = itertools.tee(
        (pd.Series(y, name="y") for y in y_data), 2
    )

    first = pd.DataFrame(x_data[0])

    col_mean = batched_numeric_col_mean(
        x=batched_x_a, y=batched_y_a, first=first, col="a"
    )
    flat_a, flat_b = (
        sum([x["a"] for x in x_data], []),
        sum([x["b"] for x in x_data], []),
    )
    assert col_mean == np.mean([f for f in flat_a if f is not None])
    col_mean = batched_numeric_col_mean(
        x=batched_x_b, y=batched_y_b, first=first, col="b"
    )
    assert col_mean == np.mean([f for f in flat_b if f is not None])


def test_empty_mean():
    x_data = [
        {"a": [None, None], "b": [3, None]},
        {"a": [None, None], "b": [None, None]},
    ]
    y_data = [[1, 2], [1, 3]]

    # duplicate iterators
    batched_x_a, batched_x_b = itertools.tee((pd.DataFrame(x) for x in x_data[1:]), 2)
    batched_y_a, batched_y_b = itertools.tee(
        (pd.Series(y, name="y") for y in y_data), 2
    )

    first = pd.DataFrame(x_data[0])

    col_mean = batched_numeric_col_mean(
        x=batched_x_a, y=batched_y_a, first=first, col="a"
    )
    assert math.isnan(col_mean)
    col_mean = batched_numeric_col_mean(
        x=batched_x_b, y=batched_y_b, first=first, col="b"
    )
    assert col_mean == 3


def test_batched_modes():
    x_data = [{"a": ["A", "A"]}, {"a": ["C", "A"]}, {"a": ["A", "B"]}]
    y_data = [[0, 0], [1, 3]]

    batched_x = (pd.DataFrame(x) for x in x_data[1:])
    batched_y = (pd.Series(y, name="y") for y in y_data)

    first = pd.DataFrame(x_data[0])

    col_mode = batched_non_object_col_mode(
        x=batched_x, y=batched_y, first=first, col="a"
    )
    assert col_mode == "A"


def test_emtpy_modes():
    x_data = [{"a": [None]}, {"a": [None, None]}, {"a": [None, None]}]
    y_data = [[0, 0], [1, 3]]

    batched_x = (pd.DataFrame(x) for x in x_data[1:])
    batched_y = (pd.Series(y, name="y") for y in y_data)

    first = pd.DataFrame(x_data[0])

    col_mode = batched_non_object_col_mode(
        x=batched_x, y=batched_y, first=first, col="a"
    )
    assert col_mode == None



import pandas as pd

from profiling.importance import varimp


def predict_float_column(df):
    """Returns float column"""
    return df["Float"].astype("float")


df = pd.DataFrame.from_dict(
    {
        "Object": [None, "", "a", "b", "c", "d", "e"],
        "Integer": [1, 2, 3, 4, 5, 6, 7],
        "Float": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        "Categorical": pd.Series(["a", "b", "c", "d", "e", "f", "g"], dtype="category"),
        "Index": ["a", "b", "c", "d", "e", "f", "g"],
    }
).set_index("Index")


def test_order():
    chart, varlist = varimp(predict_float_column, df)
    assert varlist[0] == "Float"
    assert chart.data["Variable"].iloc[0] == "Float"
    assert chart.data["Importance"].iloc[0] > 0
    assert all(chart.data["Importance"].iloc[1:] == 0.0)


def test_subset():
    profile_columns = ["Float", "Integer", "Categorical"]
    _, varlist = varimp(predict_float_column, df, profile_columns=profile_columns)
    assert "Object" not in varlist
    assert "Float" in varlist


def test_metric():
    chart, _ = varimp(predict_float_column, df, metric="Mae")
    assert chart.data["Variable"].iloc[0] == "Float"



import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

from profiling.countplot import (
    countplot_cat,
    countplot_datetime,
    countplot_num,
    get_bins,
    trim_outliers,
)


def test_get_bins():
    X_num = pd.Series(
        data=np.concatenate([np.arange(1, 26), np.arange(5, 21)]),
        dtype="Int64",
        name="num",
    )

    X_num_binned = pd.DataFrame(
        {
            "num": pd.Categorical(
                [
                    13.0,
                    10.12,
                    18.76,
                    17.8,
                    16.84,
                    15.88,
                    14.92,
                    13.96,
                    12.04,
                    11.08,
                    9.16,
                    8.2,
                    7.24,
                    6.28,
                    5.32,
                    19.72,
                    20.68,
                    21.64,
                    22.6,
                    23.56,
                    1.468,
                    2.44,
                    4.36,
                    3.4,
                    24.52,
                ],
                ordered=True,
            ),
            "Counts": np.concatenate([[2] * 16, [1] * 9]),
        }
    )

    assert_frame_equal(
        get_bins(X_num), X_num_binned, check_exact=False, check_categorical=False
    )


def test_trim_outliers():
    X_norm = pd.Series([i for i in range(1, 100)], name="var")
    outliers = pd.Series([pow(10, 4), -1 * pow(10, 4)], name="var")
    X_outliers = pd.concat([X_norm, outliers])

    X_trimmed = trim_outliers(X_outliers)
    assert_series_equal(X_trimmed, X_norm)


X = pd.DataFrame(
    {
        "num": np.concatenate([[10] * 10, [20] * 30, [30] * 40, [-10000]]),
        "cat_ordered": pd.Categorical(
            np.concatenate([[1, 2, 3, 4, 5, 6] * 13, [1, 2, 3]]),
            categories=np.arange(1, 7),
            ordered=True,
        ),
        "cat_unordered": pd.Categorical(
            np.concatenate([["a"] * 5, ["b"] * 10, ["c"] * 66]),
            categories=["b", "c", "a"],
            ordered=False,
        ),
        "datetimes": np.concatenate(
            [
                pd.date_range("2021", freq="D", periods=60).values,
                pd.date_range("2021", freq="D", periods=15).values,
                pd.date_range("2021", freq="D", periods=5).values,
                pd.date_range("2080", freq="D", periods=1).values,
            ]
        ),
    }
).astype({"num": "int64", "datetimes": "<M8[ns]"})

count_num = pd.DataFrame(
    {
        "num": pd.Categorical(
            [29.60, 20, 10.39],
            ordered=True,
        ),
        "Counts": [40, 30, 10],
    }
)  # slight change in data when calculating bin midpoint
count_cat_unordered = pd.DataFrame(
    {
        "cat_unordered": pd.Categorical(
            ["c", "b", "a"], categories=["b", "c", "a"], ordered=False
        ),
        "Counts": [66, 10, 5],
    }
)
count_cat_ordered = pd.DataFrame(
    {
        "cat_ordered": pd.Categorical(
            [1, 2, 3, 4, 5, 6],
            categories=[1, 2, 3, 4, 5, 6],
            ordered=True,
        ),
        "Counts": np.concatenate([3 * [14], 3 * [13]]),
    }
)  #  cut due to n_categories, ordering changed but corrected in final plot
count_datetime = pd.DataFrame(
    {
        "datetimes": [
            "2021-01-02T03:36:43.200000000",
            "2021-01-06T21:36:00.000000000",
            "2021-01-13T23:31:12.000000000",
            "2021-01-04T12:57:36.000000000",
            "2021-01-09T06:14:24.000000000",
            "2021-01-11T14:52:48.000000000",
            "2021-01-28T03:21:36.000000000",
            "2021-02-23T02:24:00.000000000",
            "2021-02-16T00:28:48.000000000",
            "2021-02-08T22:33:36.000000000",
            "2021-02-01T20:38:24.000000000",
            "2021-02-27T19:40:48.000000000",
            "2021-01-21T01:26:24.000000000",
            "2021-01-25T18:43:12.000000000",
            "2021-01-23T10:04:48.000000000",
            "2021-02-04T05:16:48.000000000",
            "2021-02-06T13:55:12.000000000",
            "2021-02-11T07:12:00.000000000",
            "2021-02-13T15:50:24.000000000",
            "2021-01-18T16:48:00.000000000",
            "2021-02-18T09:07:12.000000000",
            "2021-02-20T17:45:36.000000000",
            "2021-01-16T08:09:36.000000000",
            "2021-02-25T11:02:24.000000000",
            "2021-01-30T12:00:00.000000000",
        ],
        "Counts": np.concatenate([1 * [9], 3 * [6], 2 * [4], 7 * [3], 12 * [2]]),
    }
).astype({"datetimes": "<M8[ns]"})


@pytest.mark.parametrize(
    "countplot_func, x_var, count_df",
    [
        (countplot_num, X["num"], count_num),
        (countplot_cat, X["cat_ordered"], count_cat_ordered),
        (countplot_cat, X["cat_unordered"], count_cat_unordered),
        (countplot_datetime, X["datetimes"], count_datetime),
    ],
)
def test_countplot_data(countplot_func, x_var, count_df):
    count_determined = countplot_func(x_var).data.query("Counts>0")
    assert_frame_equal(
        count_determined, count_df, check_exact=False, check_categorical=False
    )


def test_nunique_num_regression():
    """Regression test for countplot nunique==1 failure case numerical columns."""
    _ = countplot_num(pd.Series([1, 1, 1, 1, 1], name="float_col"))


def test_nunique_datetime_regression():
    """Regression test for countplot nunique==1 failure case datetime columns."""
    _ = countplot_datetime(
        pd.Series(
            [
                "2021-01-01T01:00:00",
                "2021-01-01T01:00:00",
                "2021-01-01T01:00:00",
                "2021-01-01T01:00:00",
                "2021-01-01T01:00:00",
            ],
            name="datatime_col",
        ).astype("<M8[ns]")
    )



import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from tktl import Tktl

# Generate test data
from tktl.registration.validation import (
    get_input_and_output_for_profiling,
    input_pandas_representation,
    output_pandas_representation,
)

from profiling.convenience import (
    BASE_RESULTS_PATH,
    create_accuracy,
    create_countplot,
    create_dependence,
    create_expectations,
    create_importance,
    create_slice_loss,
)
from profiling.utils import vega_sanitize

X = pd.DataFrame.from_dict(
    {
        "Integer": pd.Series([1, 2, 3], dtype="int64"),
        "Float": pd.Series([None, 2.0, 3.0], dtype="float64"),
        "Object": pd.Series([None, "", "c"], dtype="object"),
        "Bool": pd.Series([True, False, True], dtype="bool"),
        "Categorical": pd.Series(["a", "b", None], dtype="category"),
        "Datetime": pd.Series(
            ["2020-01-01", None, "2020-01-03"], dtype="datetime64[ns]"
        ),
    }
)
X_unsanitized = X.rename(columns={col: col + ".Var-[Brackets]" for col in X.columns})
y_regression = pd.Series([5, 3.2, -0.2], name="Outcome")
y_binary = pd.Series([True, False, True], name="Outcome")

# Instantiate client and generate test endpoints
tktl = Tktl()


@tktl.endpoint(kind="regression", X=X, y=y_regression)
def regression(X):
    return np.random.uniform(size=len(X))


@tktl.endpoint(kind="binary", X=X, y=y_binary)
def binary(X):
    return np.random.uniform(size=len(X))


@tktl.endpoint(kind="binary", X=X, y=y_binary)
def type_sensitive(df):
    assert X.dtypes.equals(df.dtypes)  # ensure types never change during profiling
    return np.random.uniform(size=len(df))


@tktl.endpoint(kind="binary", X=X_unsanitized, y=y_binary)
def unsanitized(X):
    return np.random.uniform(size=len(X))


def test_expectations():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_expectations(
            func=endpoint.series_func,
            x=x,
            y=y,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
        )

        for var in x.columns:
            path = Path(BASE_RESULTS_PATH, name, "anatomy", "condexp", var + ".json")
            assert path.exists()
    shutil.rmtree(BASE_RESULTS_PATH)


def test_dependence():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_dependence(
            func=endpoint.series_func,
            x=x,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
        )
        for var in x.columns:
            path = Path(BASE_RESULTS_PATH, name, "anatomy", "partialdep", var + ".json")
            assert path.exists()
    shutil.rmtree(BASE_RESULTS_PATH)


def test_sliceloss():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_slice_loss(
            func=endpoint.series_func,
            x=x,
            y=y,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
            kind=endpoint.KIND.value,
        )

        for var in x.columns:
            path = Path(BASE_RESULTS_PATH, name, "anatomy", "sliceloss", var + ".json")
            assert path.exists()
    shutil.rmtree(BASE_RESULTS_PATH)


def test_countplot():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_countplot(
            x=x,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
        )

        for var in x.columns:
            print(var)
            path = Path(BASE_RESULTS_PATH, name, "anatomy", "countplot", var + ".json")
            assert path.exists()
    shutil.rmtree(BASE_RESULTS_PATH)


def test_importance():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_importance(
            func=endpoint.series_func,
            x=x,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
            kind=endpoint.KIND.value,
        )
        assert Path(BASE_RESULTS_PATH, name, "anatomy", "varimp.json").exists()
    shutil.rmtree(BASE_RESULTS_PATH)


def test_accuracy():
    for endpoint in tktl.endpoints:
        name = endpoint.func.__name__
        x_raw = input_pandas_representation(
            value=endpoint.input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=endpoint.output_schema.value, reset_index=False
        )
        x, y = get_input_and_output_for_profiling(x=x_raw, y=y_raw)
        create_accuracy(
            func=endpoint.series_func,
            x=x,
            y=y,
            endpoint_name=name,
            profile_columns=endpoint.profile_columns,
            kind=endpoint.KIND.value,
        )
        assert Path(BASE_RESULTS_PATH, name, "accuracy", "calibration.json").exists()
        assert Path(BASE_RESULTS_PATH, name, "accuracy", "metrics.json").exists()
        assert Path(BASE_RESULTS_PATH, name, "accuracy", "errors.json").exists()
        if endpoint.KIND.value == "binary":
            assert Path(BASE_RESULTS_PATH, name, "accuracy", "roc_curve.json").exists()
    shutil.rmtree(BASE_RESULTS_PATH)


@pytest.mark.parametrize(
    "var,expected",
    [(".", r"\."), ("[", r"\["), ("Var.Unsanitized[1]-.", r"Var\.Unsanitized\[1\]-\.")],
)
def test_sanitize(var, expected):
    var_sanitized = vega_sanitize(var)
    assert var_sanitized == expected
